/*
  swusb.cpp - Software USB1.1 Device Library (Bootloader for Micronucleus)
  
    Requires CPU Clock

      AVR  12.0MHz/13.5MHz and 15.0MHz-20.0MHz
      AVR1 13.5MHz/15.0MHz and 16.5MHz-20.0MHz

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "nvmctrl.h"
#include "swusb.h"

#define BOOTLOADER_ADDRESS ((uint16_t)&__vectors)
#define FLASH_WRITE_SLEEP  5
#define DEVICE_SIGNATURE   0x1e930b

#if SPM_PAGESIZE > 128
  #error "Micronucleus only supports pagesizes up to 128 bytes"
#endif

/*
 * USB Setting's
 *
 * #define USB_IO_PORT   B      // USB Port
 * #define USB_LINE_DM   3      // USB Pin (D-)
 * #define USB_LINE_DP   4      // USB Pin (D+)
 * #define USB_LED_PORT  B      // LED Port
 * #define USB_LED_PIN   1      // LED Pin
*/

#if defined(ARDUINO_AVR_DIGISPARK)
  #define USB_IO_PORT   B
  #define USB_LINE_DM   3
  #define USB_LINE_DP   4
  #define USB_LED_PORT  B
  #define USB_LED_PIN   1
#endif

#if !defined(USB_IO_PORT) || !defined(USB_LINE_DM) || !defined(USB_LINE_DP)
  #error "USB_IO_PORT, USB_LINE_DMM, USB_LINE_DP Symbol not defined."
#endif

#define AUTO_EXIT_MS     5000
#define LED_BLINK_MS     100

#define USB_SPEED        1500000UL
#define USB_PACKET_SIZE  8
#define USB_BUFFER_SIZE  (1 + USB_PACKET_SIZE + 2)
#define USB_LINE_MASK    (_BV(USB_LINE_DM) | _BV(USB_LINE_DP))

#define USB_CONCAT2(a, b)    a ## b
#define USB_CONCAT3(a, b, c) a ## b ## c
#define USB_PORT2(a, b)      USB_CONCAT2(a, b)
#define USB_PORT3(a, b, c)   USB_CONCAT3(a, b, c)

#define ASM __asm__ volatile

#if defined(VPORTA)
  #define LED_DDRPORT USB_PORT3(VPORT, USB_LED_PORT, _DIR)
  #define LED_OUTPORT USB_PORT3(VPORT, USB_LED_PORT, _OUT)
  #define LED_TGLPORT USB_PORT3(PORT , USB_LED_PORT, _OUTTGL)
  #define USB_DDRPORT USB_PORT3(VPORT, USB_IO_PORT, _DIR)
  #define USB_OUTPORT USB_PORT3(VPORT, USB_IO_PORT, _OUT)
  #define USB_TGLPORT USB_PORT3(VPORT, USB_IO_PORT, _IN ) /* not used. dummy!! */
  #define USB_INPORT  USB_PORT3(VPORT, USB_IO_PORT, _IN )
  #define OUTTGL(bit) \
  do { \
    ASM("eor  bus, " # bit); \
    ASM("out  OUTPORT, bus"); \
  } while (0)
  #define OUTTGL_ADD_CYCLE 1
#else
  #define LED_DDRPORT USB_PORT2(DDR , USB_LED_PORT)
  #define LED_OUTPORT USB_PORT2(PORT, USB_LED_PORT)
  #define LED_TGLPORT USB_PORT2(PIN , USB_LED_PORT)
  #define USB_DDRPORT USB_PORT2(DDR , USB_IO_PORT)
  #define USB_OUTPORT USB_PORT2(PORT, USB_IO_PORT)
  #define USB_TGLPORT USB_PORT2(PIN , USB_IO_PORT)
  #define USB_INPORT  USB_PORT2(PIN , USB_IO_PORT)
  #define OUTTGL(bit) \
    do { \
      ASM("out  TGLPORT, " # bit); \
    } while (0)
  #define OUTTGL_ADD_CYCLE 0
#endif

#if defined(USB_LED_PORT) && defined(USB_LED_PIN)
  #define led_begin() \
    do { \
      LED_OUTPORT &= ~_BV(USB_LED_PIN); \
      LED_DDRPORT |=  _BV(USB_LED_PIN); \
    } while (0)
  #define led_end() \
    do { \
      LED_OUTPORT &= ~_BV(USB_LED_PIN); \
      LED_DDRPORT &= ~_BV(USB_LED_PIN); \
    } while (0)
  #define led_on() \
    do { \
      LED_OUTPORT |= _BV(USB_LED_PIN); \
    } while (0)
  #define led_off() \
    do { \
      LED_OUTPORT &= ~_BV(USB_LED_PIN); \
    } while (0)
  #if defined(VPORTA)
    #define led_toggle() \
      do { \
        LED_TGLPORT = _BV(USB_LED_PIN); \
      } while (0)
  #else
    #define led_toggle() \
    do { \
      LED_TGLPORT |= _BV(USB_LED_PIN); \
    } while (0)
  #endif
#else
  #define led_begin()
  #define led_end()
  #define led_on()
  #define led_off()
  #define led_toggle()
#endif

#define CYCLE(f) ((float)F_CPU / (f))
#define ERROR(f) (int)((CYCLE(f) - (int)CYCLE(f)) * 256)

#define SETUP() \
  ASM("ldi r18, 128" ::: "r18")

#define DELAY(f, c) \
  do { \
    int delay = CYCLE(f) - (c); \
    if ((delay >= 2) && (CYCLE(f) >= 10 + OUTTGL_ADD_CYCLE)) { \
      ASM("subi r18, %0" :: "M" (ERROR(f)) : "r18"); \
      ASM("brcs .+0"); \
      delay -= 2; \
    } \
    DELAY_LOOP(delay); \
  } \
  while (0)

#define DELAY_CYCLE 3 // cycle of delay loop

#define DELAY_LOOP(c) \
  do { \
    int count = c; \
    if (count / DELAY_CYCLE) { \
      ASM("ldi r19, %0" :: "M" (count / DELAY_CYCLE) : "r19"); \
      ASM("dec r19"); \
      ASM("brne .-4"); \
    } \
    if ((count % DELAY_CYCLE) > 0) \
      ASM("nop"); \
    if ((count % DELAY_CYCLE) > 1) \
      ASM("nop"); \
  } \
  while (0)

#define SBIS_INPORT_DM(label) \
  do { \
    ASM("sbis INPORT, DM"); \
    ASM("rjmp " # label); \
  } \
  while (0)
#define SBIC_INPORT_DM(label) \
  do { \
    ASM("sbic INPORT, DM"); \
    ASM("rjmp " # label); \
  } \
  while (0)

#define REPEAT(us, code) \
  do { \
    int count = CYCLE(1000000) * us / 2; \
    if (count >  0) { code; } \
    if (count >  1) { code; } \
    if (count >  2) { code; } \
    if (count >  3) { code; } \
    if (count >  4) { code; } \
    if (count >  5) { code; } \
    if (count >  6) { code; } \
    if (count >  7) { code; } \
    if (count >  8) { code; } \
    if (count >  9) { code; } \
    if (count > 10) { code; } \
    if (count > 11) { code; } \
  } \
  while (0)

#define USB_SYNC  0b10000000

typedef enum
{
  PID_SOF      = 0b10100101,
  PID_OUT      = 0b11100001,
  PID_IN       = 0b01101001,
  PID_SETUP    = 0b00101101,
  PID_DATA0    = 0b11000011,
  PID_DATA1    = 0b01001011,
  PID_ACK      = 0b11010010,
  PID_NACK     = 0b01011010,
  PID_STALL    = 0b00011110,
} PID;

#define PID_DATA_XOR (PID_DATA0 ^ PID_DATA1)

typedef enum
{
  REQ_GET_STATUS     = 0x00, // common
  REQ_CLEAR_FEATURE  = 0x01,
  REQ_SET_FEATURE    = 0x03,
  REQ_SET_ADDRESS    = 0x05, // Standard Device Request
  REQ_GET_DESCRIPTOR = 0x06,
  REQ_SET_DESCRIPTOR = 0x07,
  REQ_GET_CONFIG     = 0x08,
  REQ_SET_CONFIG     = 0x09,
  REQ_GET_INTERFACE  = 0x0a, // Standard Interface Request
  REQ_SET_INTERFACE  = 0x0b,
} REQ;

typedef enum
{
  DES_DEVICE    = 0x01,
  DES_CONFIG    = 0x02,
  DES_STRING    = 0x03, 
  DES_INTERFACE = 0x04,
  DES_ENDPOINT  = 0x05,
} descriptor_type_t;

typedef enum
{
  RQT_RCPT_DEVICE        = 0,
  RQT_RCPT_INTERFACE     = 1,
  RQT_RCPT_ENDPOINT      = 2,
  RQT_TYPE_STANDARD      = 0,
  RQT_TYPE_CLASS         = 1,
  RQT_TYPE_VENDOR        = 2,
  RQT_DIR_HOST_TO_DEVICE = 0,
  RQT_DIR_DEVICE_TO_HOST = 1,
} RQT;

typedef enum
{
  CMD_CONNECT    = 0,
  CMD_PAGE_ADDR  = 1,
  CMD_ERASE_ALL  = 2,
  CMD_WRITE_DATA = 3,
  CMD_START_APP  = 4,
  CMD_WRITE_PAGE = 5,
} CMD;

/* USB device descriptor */
static const uint8_t DEVICE[] PROGMEM =
{
  18,                      // length of descriptor in bytes
  DES_DEVICE,              // descriptor type
  0x10, 0x01,              // USB version supported
  0xFF,                    // device class (0xFF = vendor specific class)
  0,                       // device sub class
  0,                       // protocol
  USB_PACKET_SIZE,         // max packet size
  0xD0, 0x16,              // vendor id (0x16D0 = digispark)
  0x53, 0x07,              // device id (0x0753 = digispark)
  4, 2,                    // device version (nucleus version 2.x)
  1,                       // manufacturer string index
  2,                       // product string index
  0,                       // serial number string index
  1,                       // number of configurations
};

/* USB configuration descriptor */
static const uint8_t CONFIG[] PROGMEM =
{
  9,                       // length of descriptor in bytes
  DES_CONFIG,              // descriptor type
  9 + 9, 0,                // total length of data returned (including inlined descriptors)
  1,                       // number of interfaces in this configuration
  1,                       // index of this configuration
  0,                       // configuration name string index
  0x80,                    // attributes
  100 / 2,                 // max USB current in 2mA units
  /* interface descriptor */
  9,                       // length of descriptor in bytes
  DES_INTERFACE,           // descriptor type
  0,                       // index of this interface
  0,                       // alternate setting for this interface
  0,                       // endpoints excl 0: number of endpoint descriptors to follow
  0,                       // interface class
  0,                       // interface sub class
  0,                       // interface protocol
  0,                       // string index for interface
};

/* string descriptors */
static const uint8_t STRING_LANG[] PROGMEM =
{
  1 * sizeof(wchar_t) + 2, // length of descriptor in bytes
  DES_STRING,              // descriptor type
  0x09, 0x04,              // language index (0x0409 = US-English)
};

static const uint8_t STRING_VENDOR[] PROGMEM =
{
  13 * sizeof(wchar_t) + 2, // length of descriptor in bytes
  DES_STRING,              // descriptor type
  'S', 0,
  'a', 0,
  's', 0,
  'a', 0,
  'p', 0,
  'e', 0,
  'a', 0,
  '\'', 0,
  's', 0,
  ' ', 0,
  'L', 0,
  'a', 0,
  'b', 0,
};

static const uint8_t STRING_PRODUCT[] PROGMEM =
{
  5 * sizeof(wchar_t) + 2, // length of descriptor in bytes
  DES_STRING,              // descriptor type
  's', 0,
  'w', 0,
  'u', 0,
  's', 0,
  'b', 0,
};

extern const char *__vectors; // start address of this program

static uint8_t   _address;
static uint8_t   _datapid;
static uint8_t   _rxbuf[USB_BUFFER_SIZE];
static uint8_t   _txbuf[USB_BUFFER_SIZE];
static const uint8_t *_txptr;
static uint8_t   _txcnt;
static uint8_t   _txlen;
static bool      _txpgm;
static uint8_t   _command;
static uint8_t   _ack;
static uint16_t  _pgaddr;
static uint16_t  _pgvect[2];

void SWUSB::boot(void)
{
  _pgvect[0] = pgm_read_word(0);
  _pgvect[1] = FLASHEND < 0x2000 ? 0xFFFF : pgm_read_word(2);
  led_begin();
  connect();
  if (!calibration())
    startApp();
  while (1)
  {
    poll(_rxbuf, sizeof(_rxbuf));
    startApp();
  }
}

void SWUSB::reset(void)
{
  _address = 0;
  _datapid = PID_DATA1;
  _txptr   = _txbuf;
  _txcnt   = 0;
  _txlen   = 0;
  _txpgm   = false;
  _pgaddr  = 0;
  _command = 0;
}

void SWUSB::connect(void)
{
  USB_DDRPORT &= ~USB_LINE_MASK;
  USB_OUTPORT &= ~USB_LINE_MASK;
  reset();
}

void SWUSB::disconnect(void)
{
  USB_OUTPORT &= ~USB_LINE_MASK;
  USB_DDRPORT |= USB_LINE_MASK;
  _delay_ms(10);
  USB_OUTPORT &= ~USB_LINE_MASK;
}

void SWUSB::control(const uint8_t *buf, uint8_t len)
{
  request_t *req;
  switch (buf[0])
  {
    case PID_DATA0:
    case PID_DATA1:
      _datapid = buf[0];
      if (len == 3)
        handshake(PID_ACK);
      else if (len == USB_PACKET_SIZE + 3)
      {
        req = (request_t *)(buf + 1);
        if (req->bmRequestType.type == RQT_TYPE_VENDOR)
          command(req);
        else
        {
          switch (req->bRequest)
          {
            case REQ_GET_DESCRIPTOR:
              setup(req);
              break;
            case REQ_SET_ADDRESS:
              _address = req->wValue;
              response(0, 0);
              break;
          }              
        }
      }
      break;
    case PID_IN:
      transmit();
      break;
    case PID_ACK:
      command();
      buildpacket();
      break;
  }
}

void SWUSB::setup(const request_t *req)
{
  const uint8_t *buf = 0;
  uint8_t len = 0;
  switch ((uint8_t)(req->wValue >> 8))
  {
    case DES_DEVICE: buf = DEVICE; len = sizeof(DEVICE); break;
    case DES_CONFIG: buf = CONFIG; len = sizeof(CONFIG); break;
    case DES_STRING:
      switch ((uint8_t)(req->wValue >> 0))
      {
        case 0: buf = STRING_LANG   ; len = sizeof(STRING_LANG)   ; break;
        case 1: buf = STRING_VENDOR ; len = sizeof(STRING_VENDOR) ; break;
        case 2: buf = STRING_PRODUCT; len = sizeof(STRING_PRODUCT); break;
      }
      break;
  }
  if (len)
    response(buf, len > req->wLength ? req->wLength : len, true);
}

void SWUSB::command(const request_t *req)
{
  const uint8_t *buf = 0;
  uint8_t len = 0;
  switch (_command = req->bRequest)
  {
    case CMD_CONNECT:
      static uint8_t info[6]; // nucleus version 2.x
      info[0] = (uint8_t)(BOOTLOADER_ADDRESS >> 8);
      info[1] = (uint8_t)(BOOTLOADER_ADDRESS >> 0);
      info[2] = SPM_PAGESIZE;
      info[3] = FLASH_WRITE_SLEEP | ERASE_PAGE_OPT;
      info[4] = (uint8_t)(DEVICE_SIGNATURE >> 8);
      info[5] = (uint8_t)(DEVICE_SIGNATURE >> 0);
      buf = info;
      len = sizeof(info);
      break;
    case CMD_PAGE_ADDR:
      _pgaddr = req->wIndex;
      break;
    case CMD_WRITE_DATA:
      boot_page_fill_safe(_pgaddr + 0, req->wValue);
      boot_page_fill_safe(_pgaddr + 2, req->wIndex);
      if (((_pgaddr += 4) & (SPM_PAGESIZE - 1)) == 0)
        _command = CMD_WRITE_PAGE;
      break;
  }
  response(buf, len);
}

void SWUSB::command(void)
{
  switch (_command)
  {
    case CMD_ERASE_ALL:
      _pgaddr = BOOTLOADER_ADDRESS;
      do
      {
        _pgaddr -= SPM_ERASE_PAGESIZE;
        boot_page_erase_safe(_pgaddr);
      }
      while (_pgaddr);
      boot_page_fill_safe(0, _pgvect[0]);
      boot_page_fill_safe(2, _pgvect[1]);
      boot_page_write_safe(0);
      break;
    case CMD_WRITE_PAGE:
      boot_page_write_safe(_pgaddr - SPM_PAGESIZE);
      break;
    case CMD_START_APP:
      startApp();
      break;
  }
  _command = 0;
}

void SWUSB::startApp(void)
{
  boot_rww_enable_safe();
  uint16_t opcode = pgm_read_word(BOOTLOADER_ADDRESS - 4);
  if (opcode != 0xFFFF)
  {
    disconnect();
    led_end();
    cli();
    asm volatile ("rjmp __vectors - 4"); // jump to application reset vector at end of flash
  }
}

bool SWUSB::connected(void)
{
  return _address != 0;
}

void SWUSB::response(const uint8_t *buf, uint8_t len, bool pgm)
{
  handshake(PID_ACK);
  _txptr = buf;
  _txlen = len;
  _txpgm = pgm;
  buildpacket();
}

void SWUSB::handshake(uint8_t pid)
{
  _ack = pid;
  write(&_ack, sizeof(_ack));
}

void SWUSB::transmit(void)
{
  write(_txbuf, _txcnt);
}

void SWUSB::buildpacket(void)
{
  uint8_t len = _txlen > USB_PACKET_SIZE ? USB_PACKET_SIZE : _txlen;
  _txbuf[0] = (_datapid ^= PID_DATA_XOR);
  *(uint16_t *)(_txbuf + len + 1) = crc16(_txbuf + 1, _txptr, len, _txpgm);
  _txptr += len;
  _txlen -= len;
  _txcnt  = len + 3;
  usbBusArbitration();
}

void SWUSB::poll(uint8_t *buf, uint8_t len)
{
  (void)buf; // r24:r25
  (void)len; // r22

  ASM(".equ INPORT, %0" :: "I" (_SFR_IO_ADDR(USB_INPORT)));
  ASM(".equ MASK  , %0" :: "M" (USB_LINE_MASK));
  ASM(".equ DM    , %0" :: "M" (USB_LINE_DM));
  ASM(".equ sreg  , 13" ::: "r13");
  ASM(".equ cntH  , 14" ::: "r14");
  ASM(".equ cntM  , 15" ::: "r15");
  ASM(".equ cntL  , 16" ::: "r16");
  ASM(".equ cnt0  , 17" ::: "r17");

  ASM("in   sreg, __SREG__");
  ASM("1:");
  ASM("ldi  cnt0, %0" :: "M" ((AUTO_EXIT_MS / LED_BLINK_MS) & 0xFF));
  ASM("2:");
  led_toggle();
  ASM("ldi  cntL, %0" :: "M" (((F_CPU * LED_BLINK_MS / 1000 / 9) >> 16) & 0xFF));
  ASM("mov  cntH, cntL");
  ASM("ldi  cntL, %0" :: "M" (((F_CPU * LED_BLINK_MS / 1000 / 9) >>  8) & 0xFF));
  ASM("mov  cntM, cntL");
  ASM("ldi  cntL, %0" :: "M" (((F_CPU * LED_BLINK_MS / 1000 / 9) >>  0) & 0xFF));
  ASM("rjmp 4f");
  ASM("3:" ::: "r18", "r19", "r20", "r21", "r22", "r23", "r24", "r25", "r26", "r27", "r30", "r31"); // good luck charm !!
  uint8_t n = read(buf, len);
  ASM("tst  %0" :: "r" (n));
  ASM("breq 4f");
  control(buf, n);
  ASM("4:");
  ASM("cli");
  ASM("sbis INPORT, DM");
  ASM("rjmp 3b");
  ASM("sei");
  ASM("subi cntL, 1");
  ASM("sbc  cntM, __zero_reg__");
  ASM("sbc  cntH, __zero_reg__");
  ASM("brne 4b");
  ASM("lds  r24, %0" :: "m" (_command));
  ASM("tst  r24");
  ASM("brne 1b");
  ASM("subi cnt0, 1");
  ASM("brsh 2b");
  ASM("out  __SREG__, sreg");
}

uint8_t SWUSB::read(uint8_t *buf, uint8_t len)
{
  (void)buf; // r24:r25
  (void)len; // r22

  ASM(".equ INPORT , %0" :: "I" (_SFR_IO_ADDR(USB_INPORT)));
  ASM(".equ MASK   , %0" :: "M" (USB_LINE_MASK));
  ASM(".equ DM     , %0" :: "M" (USB_LINE_DM));
  ASM(".equ DP     , %0" :: "M" (USB_LINE_DP));
  ASM(".equ bus2, 31" ::: "r31");
  ASM(".equ bus1, 30" ::: "r30");
  ASM(".equ ptrH, 27" ::: "r27");
  ASM(".equ ptrL, 26" ::: "r26");    // X register
  ASM(".equ cnt , 24" ::: "r24");
  ASM(".equ stuf, 23" ::: "r23");
  ASM(".equ len , 22" ::: "r22");    // return value
  ASM(".equ save, 21" ::: "r21");
  ASM(".equ data, 20" ::: "r20");

  // [SYNC] detect H level

  REPEAT(0.8, SBIC_INPORT_DM(101f));

  // detect reset

  DELAY(USB_SPEED, 0);
  DELAY(USB_SPEED, 0);
  DELAY(USB_SPEED, 0);
  ASM("in   bus1, INPORT");
  ASM("andi bus1, MASK");
  ASM("brne 100f");
  reset();
  ASM("100:");
  ASM("rjmp 900f"); // -> timeouted.

  // [SYNC] detect L edge

  ASM("101:");
  REPEAT(0.8, SBIS_INPORT_DM(102f));
  ASM("rjmp 900f"); // -> timeouted.

  // [SYNC] check L level

  ASM("102:");
  SETUP();
  ASM("mov  ptrL, r24");
  ASM("mov  ptrH, r25");
  ASM("clr  bus2");
  ASM("clr  save");
  DELAY(USB_SPEED, 6);
  ASM("sbic INPORT, DM");
  ASM("rjmp 101b"); // -> retry.
  ASM("clr  cnt");
  ASM("ldi  data, ~0x80");
  DELAY(USB_SPEED, 4);
  ASM("rjmp 210f"); // -> frame start

  // bit-7 unstuffing
  ASM("127:");
  ASM("ori  data, 0x40");
  ASM("andi stuf, ~0x40");
  ASM("mov  bus1, bus2");
  DELAY(USB_SPEED, 7);
  ASM("in   bus2, INPORT");
  DELAY(USB_SPEED, 8);
  ASM("rjmp 227f");

  // bit-0 unstuffing
  ASM("120:");
  ASM("ori  data, 0x80");
  ASM("mov  bus2, bus1");
  DELAY(USB_SPEED, 7);
  ASM("in   bus1, INPORT");
  DELAY(USB_SPEED, 7);
  ASM("rjmp 220f");

  // bit-1 unstuffing
  ASM("121:");
  ASM("ori  data, 0x01");
  ASM("andi stuf, ~0x01");
  ASM("mov  bus1, bus2");
  DELAY(USB_SPEED, 8);
  ASM("in   bus2, INPORT");
  DELAY(USB_SPEED, 6);
  ASM("rjmp 221f");

  // se0
  ASM("190:");
  ASM("rjmp 910f");

  // bit-7
  ASM("200:");
  ASM("in   bus2, INPORT");
  ASM("cpi  data, 0x02");
  ASM("brlo 127b");
  DELAY(USB_SPEED, 8);
  ASM("227:");
  ASM("eor  bus1, bus2");
  ASM("bst  bus1, DM");
  ASM("bld  data, 7");

  ASM("eor  stuf, data");
  ASM("mov  save, stuf");

  // bit-0
  ASM("210:");
  ASM("in   bus1, INPORT");
  ASM("ldi  stuf, 0xFF");
  ASM("andi data, 0xFC");
  ASM("breq 120b");
  DELAY(USB_SPEED, 8);
  ASM("220:");
  ASM("andi bus1, MASK");
  ASM("breq 190b");
  ASM("eor  bus2, bus1");
  ASM("bst  bus2, DM");

  // bit-1
  ASM("in   bus2, INPORT");
  ASM("bld  data, 0");    // bit-0 code
  ASM("andi data, 0xF9");
  ASM("breq 121b");
  DELAY(USB_SPEED, 7);
  ASM("221:");
  ASM("eor  bus1, bus2");
  ASM("bst  bus1, DM");
  ASM("bld  data, 1");

  // bit-2
  ASM("in   bus1, INPORT");
  ASM("andi data, 0xF3");
  ASM("breq 822f");
  DELAY(USB_SPEED, 8);
  ASM("222:");
  ASM("eor  bus2, bus1");
  ASM("bst  bus2, DM");
  ASM("bld  data, 2");

  ASM("cp   cnt, len");
  ASM("brsh 190b");

  // bit-3
  ASM("in   bus2, INPORT");
  ASM("andi data, 0xE7");
  ASM("breq 823f");
  DELAY(USB_SPEED, 8);
  ASM("223:");
  ASM("eor  bus1, bus2");
  ASM("bst  bus1, DM");
  ASM("bld  data, 3");

  ASM("st   x+, save");

  // bit-4
  ASM("in   bus1, INPORT");
  ASM("andi data, 0xCF");
  ASM("breq 824f");
  DELAY(USB_SPEED, 8);
  ASM("224:");
  ASM("eor  bus2, bus1");
  ASM("bst  bus2, DM");

  ASM("cpi  cnt, 1");
  ASM("sbc  ptrL, __zero_reg__");
  ASM("sbc  ptrH, __zero_reg__");

  // bit-5
  ASM("in   bus2, INPORT");
  ASM("bld  data, 4");         // bit-4 code
  ASM("andi data, 0x9F");
  ASM("breq 825f");
  DELAY(USB_SPEED, 8);
  ASM("225:");
  ASM("eor  bus1, bus2");
  ASM("bst  bus1, DM");
  ASM("bld  data, 5");

  ASM("inc  cnt");

  // bit-6
  ASM("in   bus1, INPORT");
  ASM("andi data, 0x3F");
  ASM("breq 826f");
  DELAY(USB_SPEED, 8);
  ASM("226:");
  ASM("eor  bus2, bus1");
  ASM("bst  bus2, DM");
  ASM("bld  data, 6");

  ASM("rjmp 200b");

  // bit-2 unstuffing
  ASM("822:");
  ASM("ori  data, 0x02");
  ASM("andi stuf, ~0x02");
  ASM("mov  bus2, bus1");
  DELAY(USB_SPEED, 7);
  ASM("in   bus1, INPORT");
  DELAY(USB_SPEED, 8);
  ASM("rjmp 222b");

  // bit-3 unstuffing
  ASM("823:");
  ASM("ori  data, 0x04");
  ASM("andi stuf, ~0x04");
  ASM("mov  bus1, bus2");
  DELAY(USB_SPEED, 7);
  ASM("in   bus2, INPORT");
  DELAY(USB_SPEED, 8);
  ASM("rjmp 223b");

  // bit-4 unstuffing
  ASM("824:");
  ASM("ori  data, 0x08");
  ASM("andi stuf, ~0x08");
  ASM("mov  bus2, bus1");
  DELAY(USB_SPEED, 7);
  ASM("in   bus1, INPORT");
  DELAY(USB_SPEED, 8);
  ASM("rjmp 224b");

  // bit-5 unstuffing
  ASM("825:");
  ASM("ori  data, 0x10");
  ASM("andi stuf, ~0x10");
  ASM("mov  bus1, bus2");
  DELAY(USB_SPEED, 7);
  ASM("in   bus2, INPORT");
  DELAY(USB_SPEED, 7);
  ASM("rjmp 225b");

  // bit-6 unstuffing
  ASM("826:");
  ASM("ori  data, 0x20");
  ASM("andi stuf, ~0x20");
  ASM("mov  bus2, bus1");
  DELAY(USB_SPEED, 7);
  ASM("in   bus1, INPORT");
  DELAY(USB_SPEED, 8);
  ASM("rjmp 226b");

  ASM("900:");
  ASM("clr  cnt");
  ASM("910:");
  ASM("cpse cnt, __zero_reg__");
  ASM("st   x+, save");

  ASM("920:");
  ASM("sbis INPORT, DM");
  ASM("rjmp 920b");
  ASM("ret");
}

void SWUSB::write(const uint8_t *buf, uint8_t len)
{
  (void)buf; // r24:r25
  (void)len; // r22

  if (len == 0)
    return;

  SETUP();

  ASM(".equ DDRPORT, %0" :: "I" (_SFR_IO_ADDR(USB_DDRPORT)));
  ASM(".equ OUTPORT, %0" :: "I" (_SFR_IO_ADDR(USB_OUTPORT)));
  ASM(".equ TGLPORT, %0" :: "I" (_SFR_IO_ADDR(USB_TGLPORT)));
  ASM(".equ INPORT , %0" :: "I" (_SFR_IO_ADDR(USB_INPORT)));
  ASM(".equ MASK   , %0" :: "M" (USB_LINE_MASK));
  ASM(".equ DM     , %0" :: "M" (USB_LINE_DM));
  ASM(".equ SYNC   , %0" :: "M" (USB_SYNC));
  ASM(".equ ptrH, 27" ::: "r27");
  ASM(".equ ptrL, 26" ::: "r26");    // X register
  ASM(".equ bus , 25" ::: "r25");
  ASM(".equ line, 24" ::: "r24");
  ASM(".equ stuf, 23" ::: "r23");
  ASM(".equ len , 22" ::: "r22");
  ASM(".equ save, 21" ::: "r21");
  ASM(".equ data, 20" ::: "r20");

  ASM("mov  ptrL, r24");
  ASM("mov  ptrH, r25");
  ASM("ldi  line, MASK");

  ASM("in   __tmp_reg__, __SREG__");
  ASM("cli");

  ASM("in   bus, OUTPORT");
  ASM("andi bus, ~MASK");
  ASM("ori  bus, 1 << DM");
  ASM("out  OUTPORT, bus");
  ASM("in   data, DDRPORT");
  ASM("or   data, line");
  ASM("out  DDRPORT, data");
  ASM("ldi  data, SYNC");
  ASM("inc  len");
  ASM("clr  stuf");
  ASM("rjmp 210f");

  // bit-7 stuffing
  ASM("127:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 227f");

  // bit-0 stuffing
  ASM("120:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 220f");

  // bit-1 stuffing
  ASM("121:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 5 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 221f");

  ASM("190:");
  ASM("rjmp 900f");

  // bit-7
  ASM("200:");
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 127b");
  DELAY(USB_SPEED, 8 + OUTTGL_ADD_CYCLE);
  ASM("227:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("brts 190b");
  ASM("mov  data, save");

  // bit-0
  ASM("210:");
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 120b");
  DELAY(USB_SPEED, 8 + OUTTGL_ADD_CYCLE);
  ASM("220:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  // bit-1
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 121b");
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("221:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  // bit-2
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 822f");
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("222:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("dec len");

  // bit-3
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 823f");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  ASM("223:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("set");

  // bit-4
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 824f");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  ASM("224:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("cpse len, __zero_reg__");
  ASM("clt");

  // bit-5
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 825f");
  DELAY(USB_SPEED, 8 + OUTTGL_ADD_CYCLE);
  ASM("225:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("ld  save, x+");   // *** Note that it is over read by 1 byte. ***

  // bit-6
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 826f");
  DELAY(USB_SPEED, 8 + OUTTGL_ADD_CYCLE);
  ASM("226:");
  ASM("lsr  data");
  ASM("ror  stuf");
  ASM("sbrs stuf, 7");
  OUTTGL(line);

  ASM("rjmp 200b");

  // bit-2 stuffing
  ASM("822:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 5 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 222b");

  // bit-3 stuffing
  ASM("823:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 223b");

  // bit-4 stuffing
  ASM("824:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 224b");

  // bit-5 stuffing
  ASM("825:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 225b");

  // bit-6 stuffing
  ASM("826:");
  ASM("clr  stuf");
  DELAY(USB_SPEED, 7 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 6 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 226b");

  // last bit stuffing
  ASM("827:");
  DELAY(USB_SPEED, 8 + OUTTGL_ADD_CYCLE);
  OUTTGL(line);
  DELAY(USB_SPEED, 4 + OUTTGL_ADD_CYCLE);
  ASM("rjmp 927f");

  ASM("900:");
  ASM("cpi  stuf, 0xFC");
  ASM("brsh 827b");
  DELAY(USB_SPEED, 8);
  ASM("927:");
  ASM("andi bus, ~MASK");
  ASM("out  OUTPORT, bus");
  DELAY(USB_SPEED, 0); // SE0 (1)
  DELAY(USB_SPEED, 3); // SE0 (2)
  ASM("in   save, DDRPORT");
  ASM("eor  save, line");
  ASM("sbi  OUTPORT, DM");
  ASM("out  DDRPORT, save");
  ASM("cbi  OUTPORT, DM");
  ASM("out  __SREG__, __tmp_reg__");
  ASM("ret");
}

/*
 * This implementation is faster, but has bigger code size
 * This version has been optimized in size compared to the original
 * fast CRC by Shay Green, January 2014.
 *
 * Thanks to Slawomir Fras (BoskiDialer) for the original contribution.
 * It implements the following C pseudo-code:
 * unsigned table(unsigned char x)
 * {
 * unsigned    value;
 *
 *     value = (unsigned)x << 6;
 *     value ^= (unsigned)x << 7;
 *     if(parity(x))
 *         value ^= 0xc001;
 *     return value;
 * }
 * unsigned usbCrc16(unsigned char *src, unsigned char len)
 * {
 * unsigned crc = 0xffff;
 *
 *     while(len--)
 *         crc = table(lo8(crc) ^ *src++) ^ hi8(crc);
 *     return ~crc;
 * }
*/
uint16_t SWUSB::crc16(uint8_t *dst, const uint8_t *src, uint8_t len, bool pgm)
{
  (void)dst; // r24:r25
  (void)src; // r22:r23
  (void)len; // r20
  (void)pgm; // r18

  ASM(".equ src    , 30" ::: "r30", "r31");  // Z register
  ASM(".equ dst    , 26" ::: "r26", "r27");  // X register
  ASM(".equ resCrcH, 25" ::: "r25");
  ASM(".equ resCrcL, 24" ::: "r24");         // return value
  ASM(".equ bitCnt , 23" ::: "r23");
  ASM(".equ scratch, 22" ::: "r22");
  ASM(".equ byte   , 21" ::: "r21");
  ASM(".equ len    , 20" ::: "r20");
  ASM(".equ copy   , 19" ::: "r19");
  ASM(".equ pgm    , 18" ::: "r18");

  ASM("movw src, r22");
  ASM("movw dst, r24");
  ASM("mov  copy, r24");
  ASM("or   copy, r25");
  ASM("ldi  resCrcL, 0xFF");
  ASM("ldi  resCrcH, 0xFF");
  ASM("clr  bitCnt");         // zero
  ASM("rjmp 9f");
  ASM("1:");
  ASM("tst  pgm");
  ASM("brne 2f");
  ASM("ld   byte, z+");
  ASM("rjmp 3f");
  ASM("2:");
  ASM("lpm  byte, z+");
  ASM("3:");
  ASM("cpse copy, __zero_reg__");
  ASM("st   x+, byte");
  ASM("eor  byte, resCrcL");  // scratch is now 'x' in table()
  ASM("mov  scratch, byte");  // compute parity of 'x'
  ASM("swap byte");
  ASM("eor  byte, scratch");
  ASM("mov  resCrcL, byte");
  ASM("lsr  byte");
  ASM("lsr  byte");
  ASM("eor  byte, resCrcL");
  ASM("inc  byte");
  ASM("andi byte, 2");        // byte is now parity(x) << 1
  ASM("cp   bitCnt, byte");   // c = (byte != 0), then put in high bit
  ASM("ror  scratch");        // so that after xoring, shifting, and xoring, it gives
  ASM("ror  byte");           // the desired 0xC0 with resCrcH
  ASM("mov  resCrcL, byte");
  ASM("eor  resCrcL, resCrcH");
  ASM("mov  resCrcH, scratch");
  ASM("lsr  scratch");
  ASM("ror  byte");
  ASM("eor  resCrcH, scratch");
  ASM("eor  resCrcL, byte");
  ASM("9:");
  ASM("subi len, 1");
  ASM("brsh 1b");
  ASM("com  resCrcL");
  ASM("com  resCrcH");
  ASM("ret");
}

void SWUSB::usbBusArbitration(void)
{
  // Bus arbitration is required because USB protocol retry enters during processing.
  ASM(".equ INPORT , %0" :: "I" (_SFR_IO_ADDR(USB_INPORT)));
  ASM(".equ MASK   , %0" :: "M" (USB_LINE_MASK));
  ASM(".equ STATE  , %0" :: "M" (1 << USB_LINE_DM));
  ASM(".equ TIMEOUT, %0" :: "M" (F_CPU * 6 / 1000000 / 7)); // 6us
  ASM(".equ loop, 25" ::: "r25");
  ASM(".equ data, 24" ::: "r24");
  ASM(".equ line, 23" ::: "r23");

  ASM("ldi  line, STATE");
  ASM("rjmp 2f");
  ASM("1:");
  ASM("in   data, INPORT");
  ASM("andi data, MASK");
  ASM("cpse data, line");
  ASM("2:");
  ASM("ldi  loop, TIMEOUT");
  ASM("subi loop, 1");
  ASM("brsh 1b");
  ASM("ret");
}

uint16_t SWUSB::usbMeasureFrameLength(void)
{
  ASM(".equ INPORT, %0" :: "I" (_SFR_IO_ADDR(USB_INPORT)));
  ASM(".equ DM    , %0" :: "M" (USB_LINE_DM));
  ASM(".equ cntH, 25" ::: "r25");
  ASM(".equ cntL, 24" ::: "r24");

  ASM("clr  cntL");
  ASM("clr  cntH");
  ASM("1:");
  ASM("adiw cntL, 1");
  ASM("breq 9f");
  ASM("sbic INPORT, DM");
  ASM("rjmp 1b");
  ASM("clr  cntL");
  ASM("clr  cntH");
  ASM("2:");
  ASM("adiw cntL, 1");
  ASM("breq 9f");
  ASM("sbis INPORT, DM");
  ASM("rjmp 2b");
  ASM("3:");
  ASM("adiw cntL, 1");
  ASM("breq 9f");
  ASM("sbic INPORT, DM");
  ASM("rjmp 3b");
  ASM("9:");
  ASM("ret");
}

#if defined(CLKCTRL_OSC20MCALIBA)
  #define GET_OSCCAL()  CLKCTRL_OSC20MCALIBA
  #define SET_OSCCAL(v) do { CPU_CCP = CCP_IOREG_gc; CLKCTRL_OSC20MCALIBA = (v); } while (0)
#else
  #define GET_OSCCAL()  OSCCAL
  #define SET_OSCCAL(v) OSCCAL = (v)
#endif

bool SWUSB::calibration(void)
{
  // USB connected ?
  uint16_t usb;
  for (uint8_t i = 0; i < 10; ++i)
  {
    if ((usb = usbMeasureFrameLength()))
      break;
  }
  if (usb)
  {
    // Calibrate OSCCAL
    uint8_t osc = GET_OSCCAL();
    uint8_t cal = osc;
    int16_t pos = 0;
    int16_t neg = 0;
    for (uint8_t i = 0; i < 8; ++i)
    {
      usb = usbMeasureFrameLength();
      if (usb == 0)
      {
        pos = neg = 0;
        break;
      }
      int16_t x = usb - (int16_t)(F_CPU * 1 / 1000 / 6 + 0.5); // 1ms
      if (x < 0)
      {
        if ((neg = -x) < pos)
          break;
        ++cal;
      }
      else
      {
        if ((pos = x) < neg)
          break;
        --cal;
      }
      SET_OSCCAL(cal);
    }
    if ((pos == 0) || (neg == 0))
      SET_OSCCAL(osc);
  }

#if 0
  // 500Hz clock for debug output
  DDRB |= _BV(PB0);
  while (1)
  {
    PINB |= _BV(PB0);
    ASM("nop");
    ASM("nop");
    ASM("nop");
    _delay_loop_2(F_CPU / 1000 / 4 - 2);
  }
#endif

  return usb;
}
